
`include "top_defines.v"
`timescale 1ns/1ps

////////////////////////////////////////////////////////////////////////////
// Width of S_AXI data bus. The slave accepts write data and issues read data
// of width C_S_AXI_DATA_WIDTH
`define    C_S_AXI_DATA_WIDTH  32
`define    C_S_AXI_ADDR_WIDTH  32
`define    C_S_AXI_REG_WIDTH 8


module `getname(top,`module_name) (
   S_AXI_ACLK,          // AXI clock
   S_AXI_ARESETN,       // AXI active low reset signal
   
                       
   S_AXI_AWADDR,        // Write address (issued by master, acceped by Slave)
   S_AXI_AWVALID,       // Write address valid. 
   S_AXI_AWREADY,       // Write address ready. 
   S_AXI_WDATA,         // Slave Interface Write Data channel Ports Write data (issued by master, acceped by Slave)
  
   S_AXI_WSTRB,         // Write strobes.
   S_AXI_WVALID,        //Write valid.
   S_AXI_WREADY,
   S_AXI_BRESP,         // Write response. 
   S_AXI_BVALID,        // Write response valid.
   S_AXI_BREADY,
   
   S_AXI_ARADDR,        // Read address (issued by master, acceped by Slave)
   S_AXI_ARVALID,       // Read address valid.
   S_AXI_ARREADY,       // Read address ready.
   S_AXI_RDATA,         //Read data (issued by slave)
   S_AXI_RRESP,         // Read response. 
   S_AXI_RVALID,        // Read valid. 
   S_AXI_RREADY,         // Read ready.
   AXI_INTA_O ,
   SCL, 
   SDA  
);

   ////////////////////////////////////////////////////////////////////////////
  // System Signals
  ////////////////////////////////////////////////////////////////////////////
  // AXI Clock Signal
  input wire S_AXI_ACLK;
  ////////////////////////////////////////////////////////////////////////////
  // AXI Active Low Reset Signal
  input wire S_AXI_ARESETN; 
  
  ////////////////////////////////////////////////////////////////////////////
  // Slave Interface Write Address channel Ports
  ////////////////////////////////////////////////////////////////////////////
  // Master Interface Write Address Channel ports
  // Write Address (Issued by Master, Acceped by Slave)
  input  wire [`C_S_AXI_ADDR_WIDTH - 1:0] S_AXI_AWADDR;
  ////////////////////////////////////////////////////////////////////////////
  // Write Address Valid. This signal indicates that the master signaling
  // valid write address and control information.
  input  wire                             S_AXI_AWVALID;
  ////////////////////////////////////////////////////////////////////////////
  // Write Address Ready. This signal indicates that the slave is ready
  // to accept an address and associated control signals.
  output wire                             S_AXI_AWREADY;

  ////////////////////////////////////////////////////////////////////////////
  // Slave Interface Write Data channel Ports
  // Write Data (Issued by Master, Acceped by Slave)
  input  wire [`C_S_AXI_DATA_WIDTH-1:0]   S_AXI_WDATA;
  ////////////////////////////////////////////////////////////////////////////
  // Write Strobes. This signal indicates which byte lanes hold
  // valid data. There is one write strobe bit for each eight
  // bits of the write data bus.
  input  wire [`C_S_AXI_DATA_WIDTH/8-1:0] S_AXI_WSTRB;
  ////////////////////////////////////////////////////////////////////////////
  //Write Valid. This signal indicates that valid write
  // data and strobes are available.
  input  wire                          S_AXI_WVALID;
  ////////////////////////////////////////////////////////////////////////////
  // Write Ready. This signal indicates that the slave
  // can accept the write data.
  output wire                          S_AXI_WREADY;

  ////////////////////////////////////////////////////////////////////////////
  // Slave Interface Write Response channel Ports
  ////////////////////////////////////////////////////////////////////////////
  // Write Response. This signal indicates the status
  // of the write transaction.
  output wire [1:0]                    S_AXI_BRESP;
  ////////////////////////////////////////////////////////////////////////////
  // Write Response Valid. This signal indicates that the channel
  // is signaling a valid write response.
  output wire                          S_AXI_BVALID;
  ////////////////////////////////////////////////////////////////////////////
  // Response Ready. This signal indicates that the master
  // can accept a write response.
  input  wire                          S_AXI_BREADY;

  ////////////////////////////////////////////////////////////////////////////
  // Slave Interface Read Address channel Ports
  // Read Address (Issued by Master, Acceped by Slave)
  input  wire [`C_S_AXI_ADDR_WIDTH - 1:0] S_AXI_ARADDR;
  ////////////////////////////////////////////////////////////////////////////
  // Read Address Valid. This signal indicates that the channel
  // is signaling valid read address and control information.
  input  wire                          S_AXI_ARVALID;
  ////////////////////////////////////////////////////////////////////////////
  // Read Address Ready. This signal indicates that the slave is
  // ready to accept an address and associated control signals.
  output wire                          S_AXI_ARREADY;

  ////////////////////////////////////////////////////////////////////////////
  // Slave Interface Read Data Channel Ports
  // Read Data (issued by slave)
  output wire [`C_S_AXI_DATA_WIDTH-1:0] S_AXI_RDATA;
  ////////////////////////////////////////////////////////////////////////////
  // Read Response. This signal indicates the status of the
  // read transfer.
  output wire [1:0]                    S_AXI_RRESP;
  ////////////////////////////////////////////////////////////////////////////
  // Read Valid. This signal indicates that the channel is
  // signaling the required read data.
  output wire                          S_AXI_RVALID;
  ////////////////////////////////////////////////////////////////////////////
  // Read Ready. This signal indicates that the master can
  // accept the read data and response information.
  input  wire                          S_AXI_RREADY;
  output reg                           AXI_INTA_O ; 
  
  inout                               SDA;
  inout                               SCL;
////////////////////////////////////////////////////////////////////////////
// Local Parameter for Addressing 32 Bit / 64 Bit C_S_AXI_DATA_WIDTH
// ADDR_LSB is Used for Addressing 32/64 Bit Registers/Memories
// ADDR_LSB = 2 for 32 bits (n downto 2)
// ADDR_LSB = 3 for 64 bits (n downto 3)

////////////////////////////////////////////////////////////////////////////
// Function called clogb2 that returns an integer which has the
// value of the ceiling of the log base 2.
  function integer clogb2 (input integer bd);
  integer bit_depth;
  begin
    bit_depth = bd;
    for(clogb2=0; bit_depth>0; clogb2=clogb2+1)
      bit_depth = bit_depth >> 1;
    end
  endfunction

  localparam integer ADDR_LSB = clogb2(`C_S_AXI_DATA_WIDTH/8)-1;
  localparam integer ADDR_MSB = `C_S_AXI_ADDR_WIDTH;

////////////////////////////////////////////////////////////////////////////
// AXI4 Lite internal signals
////////////////////////////////////////////////////////////////////////////
// read response
  reg [1 :0]                   axi_rresp;
////////////////////////////////////////////////////////////////////////////
// write response
  reg [1 :0]                   axi_bresp;
////////////////////////////////////////////////////////////////////////////
// write address acceptance
  reg                          axi_awready;
////////////////////////////////////////////////////////////////////////////
// write data acceptance
  reg                          axi_wready;
////////////////////////////////////////////////////////////////////////////
// write response valid
  reg                          axi_bvalid;
////////////////////////////////////////////////////////////////////////////
// read data valid
  reg                          axi_rvalid;
////////////////////////////////////////////////////////////////////////////
// write address
  reg     [ADDR_MSB-1:0]       axi_awaddr;
////////////////////////////////////////////////////////////////////////////
// read address valid
  reg     [ADDR_MSB-1:0]       axi_araddr;
////////////////////////////////////////////////////////////////////////////
// read data
  reg [`C_S_AXI_DATA_WIDTH-1:0] axi_rdata;
////////////////////////////////////////////////////////////////////////////
// read address acceptance
  reg                           axi_arready;
////////////////////////////////////////////////////////////////////////////
// Example-specific design signals

////////////////////////////////////////////////////////////////////////////
// Signals for user logic chip select generation

////////////////////////////////////////////////////////////////////////////
// Signals for user logic register space example
// Four slave register

////////////////////////////////////////////////////////////////////////////
// Slave Register 0
// reg [`C_S_REG_DATA_WIDTH-1:0]    slv_reg0;

  //reg  [15:0]                       prescale_reg;   // Clock Prescale Register
  reg  [ `C_S_AXI_REG_WIDTH-1:0]    control_reg;    // Control  Register
  reg  [ `C_S_AXI_REG_WIDTH-1:0]    transmit_reg;   // Transmit Register
  wire [ `C_S_AXI_REG_WIDTH-1:0]    receive_reg;    // Receive  Register
  reg  [ `C_S_AXI_REG_WIDTH-1:0]    command_reg;    // Command  Register
  wire [ `C_S_AXI_REG_WIDTH-1:0]    status_reg;     // Status   Register
  reg  [ `C_S_AXI_REG_WIDTH-1:0]    prescale_reg0;
  reg  [ `C_S_AXI_REG_WIDTH-1:0]    prescale_reg1; 
  // Done Signal: command completed, clear command register
  wire done;

 //Core Enable Signal
  wire core_en;
  wire ien;

// Status Register Signals
  wire irxack;
  reg  rxack;                 // Received Aknowledge from Slave
  reg  tip;                   // Transfer in Progress
  reg  irq_flag;              // Interrupt Pending Flag
  wire i2c_busy;             // Bus Busy (start signal detected)
  wire i2c_al;               // I2C Bus Arbitration Lost
  reg  al;                   // Status Register Arbitration Lost Bit

  // Status Signals between Byte controller and Bit controller
  wire [3:0] core_cmd;       // Output from Byte Controller to Input of Bit Controller
  wire core_txd;             // Output from Byte Controller to Input of Bit Controller
  wire core_ack; 
  wire core_rxd;             // Output from Bit Controller to Input of Byte controller

 // Assign SCL and SDA to Individual Wires
  wire scl_pad_i = SCL;
  wire scl_pad_o;       
  wire scl_padoen_o; 

  wire sda_pad_i = SDA;     
  wire sda_pad_o;       
  wire sda_padoen_o;
////////////////////////////////////////////////////////////////////////////

// Slave Register Read Enable
 wire                             slv_reg_rden;
////////////////////////////////////////////////////////////////////////////
// Slave Register Write Enable
 wire                             slv_reg_wren;
////////////////////////////////////////////////////////////////////////////
// Register Read Data
 reg [`C_S_AXI_REG_WIDTH-1:0]    reg_data_out;
 integer                          byte_index;
////////////////////////////////////////////////////////////////////////////
//I/O Connections Assignments

////////////////////////////////////////////////////////////////////////////
//Write Address Ready (AWREADY)
 assign S_AXI_AWREADY = axi_awready;

////////////////////////////////////////////////////////////////////////////
//Write Data Ready(WREADY)
 assign S_AXI_WREADY  = axi_wready;

////////////////////////////////////////////////////////////////////////////
//Write Response (BResp)and Response Valid (BVALID)
 assign S_AXI_BRESP  = axi_bresp;
 assign S_AXI_BVALID = axi_bvalid;

////////////////////////////////////////////////////////////////////////////
//Read Address Ready(AREADY)
 assign S_AXI_ARREADY = axi_arready;

////////////////////////////////////////////////////////////////////////////
//Read and Read Data (RDATA), Read Valid (RVALID) and Response (RRESP)
 assign S_AXI_RDATA  = axi_rdata;
 assign S_AXI_RVALID = axi_rvalid;
 assign S_AXI_RRESP  = axi_rresp;


////////////////////////////////////////////////////////////////////////////
// Implement Axi_awready Generation
//
//  axi_awready is asserted for one S_AXI_ACLK clock cycle when both
//  S_AXI_AWVALID and S_AXI_WVALID are asserted. axi_awready is
//  de-asserted when reset is low.

  always @( posedge S_AXI_ACLK )
  begin
    if ( S_AXI_ARESETN == 1'b0 )
      begin
        axi_awready <= 1'b0;
      end
    else
      begin
        if (~axi_awready && S_AXI_AWVALID && S_AXI_WVALID)
          begin
            ////////////////////////////////////////////////////////////////////////////
            // Slave is ready to accept write address when
            // there is a valid write address and write data
            // on the write address and data bus. This design
            // expects no outstanding transactions.
            axi_awready <= 1'b1;
          end
        else
          begin
            axi_awready <= 1'b0;
          end
      end
  end

////////////////////////////////////////////////////////////////////////////
// Implement Axi_awaddr Latching
//  This process is used to latch the address when both
//  S_AXI_AWVALID and S_AXI_WVALID are valid.

  always @( posedge S_AXI_ACLK )
  begin
    if ( S_AXI_ARESETN == 1'b0 )
      begin
        axi_awaddr <= 0;
      end
    else
      begin
        if (~axi_awready && S_AXI_AWVALID && S_AXI_WVALID)
          begin
            ////////////////////////////////////////////////////////////////////////////
            // address latching
            axi_awaddr <= S_AXI_AWADDR;
          end
      end
  end

////////////////////////////////////////////////////////////////////////////
// Implement Axi_wready Generation
//  Axi_wready is asserted for one S_AXI_ACLK clock cycle when both
//  S_AXI_AWVALID and S_AXI_WVALID are asserted. axi_wready is
//  de-asserted when reset is low.

  always @( posedge S_AXI_ACLK )
  begin
    if ( S_AXI_ARESETN == 1'b0 )
      begin
        axi_wready <= 1'b0;
      end
    else
      begin
        if (~axi_wready && S_AXI_WVALID && S_AXI_AWVALID)
          begin
            ////////////////////////////////////////////////////////////////////////////
            // SLave is ready to accept write data when
            // there is a valid write address and write data
            // on the write address and data bus. This design
            // expects no outstanding transactions.
            axi_wready <= 1'b1;
          end
        else
          begin
            axi_wready <= 1'b0;
          end
      end
  end

////////////////////////////////////////////////////////////////////////////
// Implement memory mapped register select and write logic generation
// 
// The write data is accepted and written to memory mapped
// registers (slv_reg0, slv_reg1, slv_reg2, slv_reg3) when axi_wready,
// S_AXI_WVALID, axi_wready and S_AXI_WVALID are asserted. Write strobes are used to
// select byte enables of slave registers while writing.
// These registers are cleared when reset (active low) is applied.
// 
// Slave register write enable is asserted when valid address and data are available
// and the slave is ready to accept the write address and write data.
 
assign slv_reg_wren = axi_wready && S_AXI_WVALID && axi_awready && S_AXI_AWVALID;

  always @( posedge S_AXI_ACLK )
  begin
    if ( S_AXI_ARESETN == 1'b0 )
      begin
        prescale_reg0 <= {8'b0};
        prescale_reg1 <= {8'b0};
        control_reg <= {`C_S_AXI_REG_WIDTH{1'b0}};
        transmit_reg<= {`C_S_AXI_REG_WIDTH{1'b0}};     
      end
    else begin
      if (slv_reg_wren)
        begin
          case ( axi_awaddr[ADDR_MSB-1:ADDR_LSB-2] )
            32'h0 :
              for ( byte_index = 0; byte_index <= (`C_S_AXI_DATA_WIDTH/8)-1; byte_index = byte_index+1 )
                if ( S_AXI_WSTRB[byte_index] == 1 ) begin
                  // Respective byte enables are asserted as per write strobes
                  prescale_reg0[(byte_index*8) +: 8] <= S_AXI_WDATA[(byte_index*8) +: 8];
               end
           32'h1 :
               for ( byte_index = 0; byte_index <= (`C_S_AXI_DATA_WIDTH/8)-1; byte_index = byte_index+1 )
                if ( S_AXI_WSTRB[byte_index] == 1 ) begin
                   // Respective byte enables are asserted as per write strobes
                   //slv_reg1[(byte_index*8) +: 8] <= S_AXI_WDATA[(byte_index*8) +: 8];
                   prescale_reg1[(byte_index*8) +: 8] <= S_AXI_WDATA[(byte_index*8) +: 8];
               end
            32'h2 :
              for ( byte_index = 0; byte_index <= (`C_S_AXI_DATA_WIDTH/8)-1; byte_index = byte_index+1 )
                if ( S_AXI_WSTRB[byte_index] == 1 ) begin
                  // Respective byte enables are asserted as per write strobes
                  control_reg[(byte_index*8) +: 8] <= S_AXI_WDATA[(byte_index*8) +: 8];
                end
            32'h3 :
              for ( byte_index = 0; byte_index <= (`C_S_AXI_DATA_WIDTH/8)-1; byte_index = byte_index+1 )
                if ( S_AXI_WSTRB[byte_index] == 1 ) begin
                  // Respective byte enables are asserted as per write strobes
                  transmit_reg[(byte_index*8) +: 8] <= S_AXI_WDATA[(byte_index*8) +: 8];
                end
            default : begin
                        prescale_reg0 <= prescale_reg0;
                        prescale_reg1 <= prescale_reg1;
                        control_reg   <= control_reg;
                        transmit_reg  <= transmit_reg;                  
                      end
          endcase
        end
    end
  end


  always @( posedge S_AXI_ACLK )
  begin
    if ( S_AXI_ARESETN == 1'b0 )     
        command_reg <= {`C_S_AXI_DATA_WIDTH{1'b0}};         
    else begin
      if (slv_reg_wren)
        begin
	        if (core_en & (axi_awaddr[ADDR_MSB-1:ADDR_LSB-2]== 32'h4) )
             begin
                for ( byte_index = 0; byte_index <= (`C_S_AXI_DATA_WIDTH/8)-1; byte_index = byte_index+1 )
                if ( S_AXI_WSTRB[byte_index] == 1 ) begin
	             command_reg [(byte_index*8) +: 8]<= S_AXI_WDATA[(byte_index*8) +: 8];
                         end
             end
	    end
	  else
	    begin
	        if (done | i2c_al)
	          command_reg[7:4] <=  4'h0;             // clear command bits when done            
	                                                   // or when aribitration lost
	          command_reg[2:1] <=  2'b00;             // reserved bits
	          command_reg[0]   <=  1'b0;             // clear IRQ_ACK bit
	    end
    end
   end
////////////////////////////////////////////////////////////////////////////
// Implement Write Response Logic Generation
//
//  The write response and response valid signals are asserted by the slave
//  when axi_wready, S_AXI_WVALID, axi_wready and S_AXI_WVALID are asserted.
//  This marks the acceptance of address and indicates the status of
//  write transaction.

  always @( posedge S_AXI_ACLK )
  begin
    if ( S_AXI_ARESETN == 1'b0 )
      begin
        axi_bvalid  <= 0;
        axi_bresp   <= 2'b00;
      end
    else
      begin
        if (axi_awready && S_AXI_AWVALID && ~axi_bvalid && axi_wready && S_AXI_WVALID)
          begin
            // Indicates a valid write response is available
            axi_bvalid <= 1'b1;
            axi_bresp  <= 2'b0; // 'OKAY' response
          end                   // work error responses in future
        else
          begin
            if (S_AXI_BREADY && axi_bvalid)
              //Check if bready is asserted while bvalid is high)
              //(there is a possibility that bready is always asserted high)
              begin
                axi_bvalid <= 1'b0;
              end
          end
      end
  end


////////////////////////////////////////////////////////////////////////////
// Implement Axi_arready Generation
//
//  Axi_arready is asserted for one S_AXI_ACLK clock cycle when S_AXI_ARVALID 
//  is asserted. axi_awready is de-asserted when reset (active low) is asserted.
//  The read address is also latched when S_AXI_ARVALID is
//  asserted. axi_araddr is reset to zero on reset assertion.

  always @( posedge S_AXI_ACLK )
  begin
    if ( S_AXI_ARESETN == 1'b0 )
      begin
        axi_arready <= 1'b0;
        axi_araddr  <= {ADDR_MSB{1'b0}};
      end
    else
      begin
        if (~axi_arready && S_AXI_ARVALID)
          begin
            // Indicates that the slave has acceped the valid read address
            axi_arready <= 1'b1;
            axi_araddr  <= S_AXI_ARADDR;
          end
        else
          begin
            axi_arready <= 1'b0;
          end
      end
  end

////////////////////////////////////////////////////////////////////////////
// Implement Memory Mapped Register Select and Read Logic Generation
//
//  axi_rvalid is asserted for one S_AXI_ACLK clock cycle when both
//  S_AXI_ARVALID and axi_arready are asserted. The slave registers
//  data are available on the axi_rdata bus at this instance. The
//  assertion of axi_rvalid marks the validity of read data on the
//  bus and axi_rresp indicates the status of read transaction.axi_rvalid
//  is deasserted on reset (active low). axi_rresp and axi_rdata are
//  cleared to zero on reset (active low).

  always @( posedge S_AXI_ACLK )
  begin
    if ( S_AXI_ARESETN == 1'b0 )
      begin
        axi_rvalid <= 0;
        axi_rresp  <= 0;
      end
    else
      begin
        if (axi_arready && S_AXI_ARVALID && ~axi_rvalid)
          begin
            // Valid read data is available at the read data bus
            axi_rvalid <= 1'b1;
            axi_rresp  <= 2'b0; // 'OKAY' response
          end
        else if (axi_rvalid && S_AXI_RREADY)
          begin
            // Read data is accepted by the master
            axi_rvalid <= 1'b0;
          end
      end
  end
  
////////////////////////////////////////////////////////////////////////////
// Slave register read enable is asserted when valid address is available
// and the slave is ready to accept the read address.
  assign slv_reg_rden = axi_arready & S_AXI_ARVALID & ~axi_rvalid;

  always @( prescale_reg0, prescale_reg1,control_reg,receive_reg,status_reg,transmit_reg,
            command_reg, S_AXI_ARESETN, slv_reg_rden, axi_araddr)
  begin
    if ( S_AXI_ARESETN == 1'b0 )
      begin
        reg_data_out <= {`C_S_AXI_DATA_WIDTH{1'b0}};
      end
    else
      begin
        // Read address mux
        case ( axi_araddr [ADDR_MSB-1:ADDR_LSB-2])
           32'h0: reg_data_out <=  prescale_reg0;
	       32'h1: reg_data_out <=  prescale_reg1;
	       32'h2: reg_data_out <=  control_reg;
	       32'h3: reg_data_out <=  receive_reg;    // write is transmit register (txr)
	       32'h4: reg_data_out <=  status_reg;     // write is command register (cr)
	       //3'h5: reg_data_out <=  transmit_reg;
	      // 3'h6: reg_data_out <=  command_reg;
	      // 3'h7: reg_data_out <=  0;               // reserved        
         default : reg_data_out <= {`C_S_AXI_DATA_WIDTH{1'b0}};
        endcase
        //end
      end
  end

  always @( posedge S_AXI_ACLK )
  begin
    if ( S_AXI_ARESETN == 1'b0 )
      begin
        axi_rdata  <= 0;
      end
    else
      begin
        ////////////////////////////////////////////////////////////////////////////
        // When there is a valid read address (S_AXI_ARVALID) with
        // acceptance of read address by the slave (axi_arready),
        // output the read dada
        if (axi_arready && S_AXI_ARVALID && ~axi_rvalid)
          begin
            axi_rdata <= reg_data_out;     // register read data
          end
      end
  end

 ////////////////////////////////////////////////////////////////////////////////////
  // decode command register
	wire sta  = command_reg[7];
	wire sto  = command_reg[6];
	wire rd   = command_reg[5];
	wire wr   = command_reg[4];
	wire ack  = command_reg[3];
	wire iack = command_reg[0];

 // decode control register
	assign core_en = control_reg[7];
	assign ien = control_reg[6];
/////////////////////////////////////////////////////////////////
// status register block + interrupt request signal
  always @(posedge S_AXI_ACLK)
  begin
    if ( S_AXI_ARESETN == 1'b0)
	    begin
	        al       <= 1'b0;
	        rxack    <= 1'b0;
	        tip      <= 1'b0;
	        irq_flag <= 1'b0;
	    end
	else
	    begin
	        al       <= i2c_al | (al & ~sta);
	        rxack    <= irxack;
	        tip      <= (rd | wr);
	        irq_flag <= (done | i2c_al | irq_flag) & ~iack; // interrupt request flag is
                                                            // always generated
	    end
end

// generate interrupt request signals
 always @(posedge S_AXI_ACLK)
  begin
    if ( S_AXI_ARESETN == 1'b0)
	    AXI_INTA_O <=  1'b0;
	   else
	    AXI_INTA_O <=  irq_flag && ien; // Interrupt signal is only generated
                                        //when IEN (interrupt enable bit is set)
  end
 
	// assign status register bits
	assign status_reg[7]   = rxack;
	assign status_reg[6]   = i2c_busy;
	assign status_reg[5]   = al;
	assign status_reg[4:2] = 3'h0; // reserved
	assign status_reg[1]   = tip;
	assign status_reg[0]   = irq_flag;	

///////////////////////////////////////////
	// Hookup Byte Controller Block
 `getname(I2C_Master_Byte_Ctrl,`module_name)   byte_controller (
		.CLK      ( S_AXI_ACLK  ),
		.NRESET   ( S_AXI_ARESETN ),
		.CLK_CNT  ( {prescale_reg1,prescale_reg0} ),
		.START    ( sta          ),
		.STOP     ( sto          ),
		.READ     ( rd           ),
		.WRITE    ( wr           ),
		.ACK_IN   ( ack          ),//command_reg[3]
		.DIN      ( transmit_reg ),
		.CMD_ACK  ( done         ),
		.ACK_OUT  ( irxack       ),//status_reg[7]
		.DOUT     ( receive_reg  ),
		.I2C_AL   ( i2c_al       ),
		.CORE_CMD ( core_cmd     ),
		.CORE_ACK ( core_ack      ),
		.CORE_TXD ( core_txd     ),
		.CORE_RXD ( core_rxd     )
	);
	
   `getname(I2C_Master_Bit_Ctrl,`module_name)   bit_controller (
		.CLK     ( S_AXI_ACLK),
		.NRESET  (  S_AXI_ARESETN),
		.ENA     ( core_en  ),
		.CLK_CNT ({prescale_reg1[7:0],prescale_reg0[7:0]}),
		.CMD     ( core_cmd ),
		.CMD_ACK ( core_ack ),
		.BUSY    ( i2c_busy ),
		.AL      ( i2c_al   ),          // output to other modules
		.DIN     ( core_txd ),
		.DOUT    ( core_rxd ),
		.SCL_I   ( scl_pad_i),
		.SCL_O   ( scl_pad_o),
		.SCL_OEN ( scl_padoen_o ),
		.SDA_I   ( sda_pad_i),
		.SDA_O   ( sda_pad_o),
		.SDA_OEN ( sda_padoen_o  )
	);
    assign SCL=(scl_padoen_o ? 1'bz : scl_pad_o);
    assign SDA=(sda_padoen_o ? 1'bz : sda_pad_o);
endmodule


module `getname(I2C_Master_Bit_Ctrl,`module_name)   (
	   CLK,
	   NRESET,
	   CLK_CNT,           // Clock Prescale Value
	   ENA,               // Core Enable Signal
	   CMD,
	   CMD_ACK,          // Command Complete Acknowledge
	   BUSY,             // I2C Bus Busy
	   AL ,              // I2C Bus Arbitration Lost
	   DIN,
	   DOUT,
	   SCL_I,           // I2C Clock Line Input
	   SCL_O,           // I2C Clock Line Output
	   SCL_OEN,         // I2I Clock Line Output Enable (Active Low)
	   SDA_I,           // I2C Data Line Input
	   SDA_O,           // I2C Data Line Output
	   SDA_OEN          // I2C Data Line Output Enable (Active Low)
	);

	
	// Inputs & Outputs
	
	input  CLK;
	input   NRESET;
	input   ENA;            
	input [15:0] CLK_CNT; 

	input [3:0] CMD;
	output      CMD_ACK; 	
	output      BUSY;    	
	output      AL;      
	
	input       DIN;
	output      DOUT;
	

	// I2C Lines
	input  SCL_I;         
	output SCL_O;         
	output SCL_OEN;       
	
	input  SDA_I;         
	output SDA_O;         
	output SDA_OEN;       

    reg CMD_ACK;
    reg BUSY;
    reg AL;
    reg DOUT;
    reg SCL_OEN;
    reg SDA_OEN;
	
	// Variable Declarations

	reg sSCL, sSDA;              // Synchronized SCL and SDA Inputs
	reg dscl_oen;                // Delayed Scl_oen
	reg sda_chk;                 // Check SDA Output (Multi-master arbitration)
	reg clk_en;                  // Clock Generation Signals
	wire slave_wait;
   //reg [15:0] cnt = clk_cnt;   // Clock Divider Counter (Simulation)
    reg [15:0] cnt;              // Clock Divider Counter (Synthesis)

	// State Machine Variable
	reg [16:0] c_state; /*// synopsys enum_state */
	
	// Module Body
	
	// Whenever the slave is not ready it can delay the cycle by pulling SCL low
	// Delay SCL_OEN
	always @(posedge CLK)
	  dscl_oen <= #1 SCL_OEN;

	assign slave_wait = dscl_oen && !sSCL;

	// Generate Clk Enable Signal
	always @(posedge CLK)
	  if(~NRESET)
	    begin
	        cnt    <= #1 16'h0;
	        clk_en <= #1 1'b1;
	    end
	   else if ( ~|cnt || !ENA)
	     begin
	        cnt    <= #1 CLK_CNT;
	        clk_en <= #1 1'b1;
	     end
	  else if (slave_wait)
	    begin
	        cnt    <= #1 cnt;
	        clk_en <= #1 1'b0;    
	    end
	  else
	    begin
	        cnt    <= #1 cnt - 16'h1;
	        clk_en <= #1 1'b0;
	    end


	// Generate Bus Status Controller
	reg dSCL, dSDA;
	reg sta_condition;
	reg sto_condition;

	// Synchronize SCL and SDA Inputs
	// Reduce Metastability Risc
	always @(posedge CLK)
	  if (~NRESET)
	    begin
	        sSCL <= #1 1'b1;
	        sSDA <= #1 1'b1;

	        dSCL <= #1 1'b1;
	        dSDA <= #1 1'b1;
	    end
	 else
	    begin
	        sSCL <= #1 SCL_I ;
	        sSDA <= #1 SDA_I;

	        dSCL <= #1 sSCL;
	        dSDA <= #1 sSDA;
	    end

	// Detect Start Condition => detect falling edge on SDA while SCL is high
	// Detect Stop Condition => detect rising edge on SDA while SCL is high
	always @(posedge CLK)
	  if (~NRESET)
	    begin
	        sta_condition <= #1 1'b0;
	        sto_condition <= #1 1'b0;
	    end
	 else
	    begin
	        sta_condition <= #1 ~sSDA &  dSDA & sSCL;
	        sto_condition <= #1  sSDA & ~dSDA & sSCL;
	    end

	// Generate I2C Bus Busy Signal
	always @(posedge CLK)
	  if(!NRESET)
	    BUSY <= #1 1'b0;
	  else
	    BUSY <= #1 (sta_condition | BUSY ) & ~sto_condition;

	// Generate Arbitration Lost Signal
	// aribitration lost when:
	// 1) master drives SDA high, but the i2c bus is low
	// 2) stop detected while not requested
	
	reg cmd_stop;
	always @(posedge CLK)
	  if (~NRESET)
	    cmd_stop <= #1 1'b0;
      else if (clk_en)
	    cmd_stop <= #1 CMD == `I2C_CMD_STOP;

	always @(posedge CLK)
	  if (~NRESET)
	    AL <= #1 1'b0;
	  else
	    AL <= #1 (sda_chk & ~sSDA & SDA_OEN) | (|c_state & sto_condition & ~cmd_stop);


	// Generate Dout Signal (store SDA on rising edge of SCL)
	always @(posedge CLK)
	  if(sSCL & ~dSCL)
	    DOUT <= #1 sSDA;

	// Generate Statemachine

	// Nxt_state Decoder
	parameter [16:0] idle    = 17'b0_0000_0000_0000_0000;
	parameter [16:0] start_a = 17'b0_0000_0000_0000_0001;
	parameter [16:0] start_b = 17'b0_0000_0000_0000_0010;
	parameter [16:0] start_c = 17'b0_0000_0000_0000_0100;
	parameter [16:0] start_d = 17'b0_0000_0000_0000_1000;
	parameter [16:0] start_e = 17'b0_0000_0000_0001_0000;
	parameter [16:0] stop_a  = 17'b0_0000_0000_0010_0000;
	parameter [16:0] stop_b  = 17'b0_0000_0000_0100_0000;
	parameter [16:0] stop_c  = 17'b0_0000_0000_1000_0000;
	parameter [16:0] stop_d  = 17'b0_0000_0001_0000_0000;
	parameter [16:0] rd_a    = 17'b0_0000_0010_0000_0000;
	parameter [16:0] rd_b    = 17'b0_0000_0100_0000_0000;
	parameter [16:0] rd_c    = 17'b0_0000_1000_0000_0000;
	parameter [16:0] rd_d    = 17'b0_0001_0000_0000_0000;
	parameter [16:0] wr_a    = 17'b0_0010_0000_0000_0000;
	parameter [16:0] wr_b    = 17'b0_0100_0000_0000_0000;
	parameter [16:0] wr_c    = 17'b0_1000_0000_0000_0000;
	parameter [16:0] wr_d    = 17'b1_0000_0000_0000_0000;

	always @(posedge CLK)
	  if (!NRESET)
	    begin
	        c_state <= #1 idle;
	        CMD_ACK <= #1 1'b0;
	        SCL_OEN <= #1 1'b1;
	        SDA_OEN <= #1 1'b1;
	        sda_chk <= #1 1'b0;
	    end
	  else if (AL)
	    begin
	        c_state <= #1 idle;
	        CMD_ACK <= #1 1'b0;
	        SCL_OEN <= #1 1'b1;
	        SDA_OEN <= #1 1'b1;
	        sda_chk <= #1 1'b0;
	    end
	  else
	    begin
	      CMD_ACK <= #1 1'b0;   // Default no command acknowledge + assert cmd_ack only 1clk cycle
	     if (clk_en)
	     case (c_state)       // synopsys full_case parallel_case
	            // Idle State
	          idle:
	          begin
	            case (CMD)    // synopsys full_case parallel_case
	                `I2C_CMD_START:
	                   c_state <= #1 start_a;
	                `I2C_CMD_STOP:
	                   c_state <= #1 stop_a;
	                `I2C_CMD_WRITE:
	                   c_state <= #1 wr_a;
	                `I2C_CMD_READ:
	                   c_state <= #1 rd_a;
	                 default:
	                   c_state <= #1 idle;
	           endcase
	                SCL_OEN <= #1 SCL_OEN; // Keep SCL in same state
	                SDA_OEN <= #1 SDA_OEN; // Keep SDA in same state
	                sda_chk <= #1 1'b0;    // Don't check SDA output
	          end
	            // Start
	         start_a:
	          begin
	             c_state <= #1 start_b;
	             SCL_OEN <= #1 SCL_OEN; // Keep SCL in same state
	             SDA_OEN <= #1 1'b1;    // Set SDA high
	             sda_chk <= #1 1'b0;    // Don't check SDA output
	          end
	        
            start_b:
	          begin
	             c_state <= #1 start_c;
	             SCL_OEN <= #1 1'b1; // set SCL high
	             SDA_OEN <= #1 1'b1; // keep SDA high
	             sda_chk <= #1 1'b0; // don't check SDA output
	          end

	        start_c:
	          begin
	            c_state <= #1 start_d;
	            SCL_OEN <= #1 1'b1; // keep SCL high
	            SDA_OEN <= #1 1'b0; // set SDA low
	            sda_chk <= #1 1'b0; // don't check SDA output
	          end

	        start_d:
	          begin
	            c_state <= #1 start_e;
	            SCL_OEN <= #1 1'b1; // keep SCL high
	            SDA_OEN <= #1 1'b0; // keep SDA low
	            sda_chk <= #1 1'b0; // don't check SDA output
	          end

	        start_e:
	          begin
	            c_state <= #1 idle;
	            CMD_ACK <= #1 1'b1;
	            SCL_OEN <= #1 1'b0; // Set SCL low
	            SDA_OEN <= #1 1'b0; // Keep SDA low
	            sda_chk <= #1 1'b0; // Don't check SDA output
	          end
	         //Stop
	        stop_a:
	          begin
	            c_state <= #1 stop_b;
	            SCL_OEN <= #1 1'b0; // Keep SCL low
	            SDA_OEN <= #1 1'b0; // Set SDA low
	            sda_chk <= #1 1'b0; // Don't check SDA output
	          end

	        stop_b:
	         begin
	            c_state <= #1 stop_c;
	            SCL_OEN <= #1 1'b1; // Set SCL high
	            SDA_OEN <= #1 1'b0; // Keep SDA low
	            sda_chk <= #1 1'b0; // Don't check SDA output
	         end

	        stop_c:
	         begin
	            c_state <= #1 stop_d;
	            SCL_OEN <= #1 1'b1; // Keep SCL high
	            SDA_OEN <= #1 1'b0; // Keep SDA low
	            sda_chk <= #1 1'b0; // Don't check SDA output
	         end

	        stop_d:
	          begin
	            c_state <= #1 idle;
	            CMD_ACK <= #1 1'b1;
	            SCL_OEN <= #1 1'b1; // Keep SCL high
	            SDA_OEN <= #1 1'b1; // Set SDA high
	            sda_chk <= #1 1'b0; // Don't check SDA output
	          end

	        // Read
	        rd_a:
	          begin
	            c_state <= #1 rd_b;
	            SCL_OEN <= #1 1'b0; // Keep SCL low
	            SDA_OEN <= #1 1'b1; // Tri-state SDA
	            sda_chk <= #1 1'b0; // Don't check SDA output
	          end

	        rd_b:
	          begin
	            c_state <= #1 rd_c;
	            SCL_OEN <= #1 1'b1; // Set SCL high
	            SDA_OEN <= #1 1'b1; // Keep SDA tri-stated
	            sda_chk <= #1 1'b0; // Don't check SDA output
	          end

	        rd_c:
	          begin
	            c_state <= #1 rd_d;
	            SCL_OEN <= #1 1'b1; // Keep SCL high
	            SDA_OEN <= #1 1'b1; // Keep SDA tri-stated
	            sda_chk <= #1 1'b0; // Don't check SDA output
	          end

	        rd_d:
	          begin
	            c_state <= #1 idle;
	            CMD_ACK <= #1 1'b1;
	            SCL_OEN <= #1 1'b0; // Set SCL low
	            SDA_OEN <= #1 1'b1; // Keep SDA tri-stated
	            sda_chk <= #1 1'b0; // Don't check SDA output
	          end

	        // Write
	        wr_a:
	          begin
	            c_state <= #1 wr_b;
	            SCL_OEN <= #1 1'b0; // Keep SCL low
	            SDA_OEN <= #1 DIN;  // Set SDA
	            sda_chk <= #1 1'b0; // Don't check SDA output (SCL low)
	          end

	        wr_b:
	          begin
	            c_state <= #1 wr_c;
	            SCL_OEN <= #1 1'b1; // Set SCL high
	            SDA_OEN <= #1 DIN; // Keep SDA
	            sda_chk <= #1 1'b1; // Check SDA output
	          end
	           
            wr_c:
	          begin
	            c_state <= #1 wr_d;
	            SCL_OEN <= #1 1'b1; // Keep SCL high
	            SDA_OEN <= #1 DIN;
	            sda_chk <= #1 1'b1; // Check SDA output
	          end

	        wr_d:
	          begin
	            c_state <= #1 idle;
	            CMD_ACK <= #1 1'b1;
	            SCL_OEN <= #1 1'b0; // Set SCL low
	            SDA_OEN <= #1 DIN;
	            sda_chk <= #1 1'b0; // Don't check SDA output (SCL low)
	          end

	  endcase
	  end


	// Assign Scl and Sda Output (always gnd)
	assign SCL_O = 1'b0;
	assign SDA_O = 1'b0;

endmodule


module `getname(I2C_Master_Byte_Ctrl,`module_name) (
	   CLK,                  // Master Clock
	   NRESET,               // Synchronous Active Low Reset
	   //ENA, 
	   CLK_CNT, 
	   START, 
	   STOP, 
	   READ, 
	   WRITE, 
	   ACK_IN, 
	   DIN,
	   CMD_ACK, 
	   ACK_OUT, 
	   DOUT, 
	   //i2c_busy, 
  	   I2C_AL,              // I2C_AL is Output From Bit Cntrl, but Input To Byte Cntrl Now
	   CORE_CMD, 
	   CORE_TXD, 
	   CORE_RXD, 
	   CORE_ACK
     );

	
	// Inputs & Outputs
	input CLK;         
	input NRESET;  
	//input ena;     // core enable signal

	input [15:0] CLK_CNT; // 4x SCL

	// Control Inputs
	input       START;
	input       STOP;
	input       READ;
	input       WRITE;
	input       ACK_IN;
	input [7:0] DIN;

	// Status Outputs
	output       CMD_ACK;	
	output       ACK_OUT;
	
	//Output       I2C_BUSY;
	input       I2C_AL;  
	output [7:0] DOUT;

	// Signals For Bit_Controller
	output  [3:0] CORE_CMD;
	output        CORE_TXD;
	input         CORE_ACK;
	input         CORE_RXD;
    reg           CMD_ACK;
    reg           ACK_OUT;
	
	// Variable Declarations
	
	// Statemachine
	parameter [4:0] ST_IDLE  = 5'b0_0000;
	parameter [4:0] ST_START = 5'b0_0001;
	parameter [4:0] ST_READ  = 5'b0_0010;
	parameter [4:0] ST_WRITE = 5'b0_0100;
	parameter [4:0] ST_ACK   = 5'b0_1000;
	parameter [4:0] ST_STOP  = 5'b1_0000;

	// Signals for Bit_controller
	reg  [3:0] CORE_CMD;
	reg        CORE_TXD;

	// Signals for Shift Register
	reg [7:0] sr;            //8Bit Shift Register
	reg       shift, ld;

	// Signals for State Machine
	wire       go;
	reg  [2:0] dcnt;
	wire       cnt_done;

	// Module body

	// Generate Go-Signal
	assign go = (READ | WRITE | STOP) & ~CMD_ACK;

	// Assign Dout Output to Shift-Register
	assign DOUT = sr;

	// Generate Shift Register
	always @(posedge CLK)
	  if (!NRESET)
	    sr <= #1 8'h0;
	  else if (ld)
	    sr <= #1 DIN;
	  else if (shift)
	    sr <= #1 {sr[6:0], CORE_RXD};

	// generate counter
	always @(posedge CLK)
	  if (!NRESET)
	    dcnt <= #1 3'h0;
	  else if (ld)
	    dcnt <= #1 3'h7;
	  else if (shift)
	    dcnt <= #1 dcnt - 3'h1;

	assign cnt_done = ~(|dcnt);

	//
	// State Machine
	//
	reg [4:0] c_state; // synopsis enum_state

	always @(posedge CLK)
	  if (!NRESET)
	    begin
	        CORE_CMD <= #1 `I2C_CMD_NOP;
	        CORE_TXD <= #1 1'b0;
	        shift    <= #1 1'b0;
	        ld       <= #1 1'b0;
	        CMD_ACK  <= #1 1'b0;
	        c_state  <= #1 ST_IDLE;
	        ACK_OUT  <= #1 1'b0;
	    end
	  else if (I2C_AL)
	   begin
	       CORE_CMD <= #1 `I2C_CMD_NOP;
	       CORE_TXD <= #1 1'b0;
	       shift    <= #1 1'b0;
	       ld       <= #1 1'b0;
	       CMD_ACK  <= #1 1'b0;
	       c_state  <= #1 ST_IDLE;
	       ACK_OUT  <= #1 1'b0;
	   end
	else
	  begin
	      // Initially Reset All Signals
	      CORE_TXD <= #1 sr[7];
	      shift    <= #1 1'b0;
	      ld       <= #1 1'b0;
	      CMD_ACK  <= #1 1'b0;

	      case (c_state) // synopsys full_case parallel_case
	        ST_IDLE:
	          if (go)
	            begin
	                if (START)
	                  begin
	                      c_state  <= #1 ST_START;
	                      CORE_CMD <= #1 `I2C_CMD_START;
	                  end
	                else if (READ)
	                  begin
	                      c_state  <= #1 ST_READ;
	                      CORE_CMD <= #1 `I2C_CMD_READ;
	                  end
	                else if (WRITE)
	                  begin
	                      c_state  <= #1 ST_WRITE;
	                      CORE_CMD <= #1 `I2C_CMD_WRITE;
	                  end
	                else // stop
	                  begin
	                      c_state  <= #1 ST_STOP;
	                      CORE_CMD <= #1 `I2C_CMD_STOP;
	                  end

	                ld <= #1 1'b1;
	            end

	        ST_START:
	          if (CORE_ACK)
	            begin
	                if (READ)
	                  begin
	                      c_state  <= #1 ST_READ;
	                      CORE_CMD <= #1 `I2C_CMD_READ;
	                  end
	                else
	                  begin
	                      c_state  <= #1 ST_WRITE;
	                      CORE_CMD <= #1 `I2C_CMD_WRITE;
	                  end

	                ld <= #1 1'b1;
	            end

	        ST_WRITE:
	          if (CORE_ACK)
	            if (cnt_done)
	              begin
	                  c_state  <= #1 ST_ACK;
	                  CORE_CMD <= #1 `I2C_CMD_READ;
	              end
	            else
	              begin
	                  c_state  <= #1 ST_WRITE;       // stay in same state
	                  CORE_CMD <= #1 `I2C_CMD_WRITE; // write next bit
	                  shift    <= #1 1'b1;
	              end

	        ST_READ:
	          if (CORE_ACK)
	            begin
	                if (cnt_done)
	                  begin
	                      c_state  <= #1 ST_ACK;
	                      CORE_CMD <= #1 `I2C_CMD_WRITE;
	                  end
	                else
	                  begin
	                      c_state  <= #1 ST_READ;       // stay in same state
	                      CORE_CMD <= #1 `I2C_CMD_READ; // read next bit
	                  end

	                shift    <= #1 1'b1;
	                CORE_TXD <= #1 ACK_IN;
	            end

	        ST_ACK:
	          if (CORE_ACK)
	            begin
	               if (STOP)
	                 begin
	                     c_state  <= #1 ST_STOP;
	                     CORE_CMD <= #1 `I2C_CMD_STOP;
	                 end
	               else
	                 begin
	                     c_state  <= #1 ST_IDLE;
	                     CORE_CMD <= #1 `I2C_CMD_NOP;

	                     // Generate Command Acknowledge Signal
	                     CMD_ACK  <= #1 1'b1;
	                 end
	                 // Assign Ack_out Output to Bit_Controller_Rxd (Contains Last Received Bit)
	                 ACK_OUT <= #1 CORE_RXD;
	                 CORE_TXD <= #1 1'b1;
	             end
	           else
	             CORE_TXD <= #1 ACK_IN;

	        ST_STOP:
	          if (CORE_ACK)
	            begin
	                c_state  <= #1 ST_IDLE;
	                CORE_CMD <= #1 `I2C_CMD_NOP;
	                // Generate Command Acknowledge Signal
	                CMD_ACK  <= #1 1'b1;
	            end

	      endcase
	  end
endmodule

